/*
 * Copyright (c) 2009, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "strictmalloc.h"
#include "expression.h"
#include "conf_parser.h"
#include "confnode.h"
#include "conf_htable.h"
#include "parser_common.h"
#include "dotconfig.h"


/*
 * Add a token in the line structure given as argument.
 *
 * 'linestart': pointer to the line buffer.
 * 'linecur': current pointer in the line string (same string than linestart).
 *
 * Return a pointer to the first token copied from buf in a allocated
 * buffer which is correctly nil-terminated. Also fill *eatlen
 * argument the length of eaten bytes. The user should free the token
 * structure and the string after use. On error, return NULL.
 */
static const char *append_token(struct line *line, const char *linestart,
				const char *linecur)
{
	struct token *tok;
	const char *start, *s;
	unsigned len;
	char *retbuf;

	tok = strictmalloc(sizeof(struct token));

	/* skip spaces */
	s = linecur;
	while (*s != '\0' && isspace(*s))
		s++;
	if (*s == '\0' || *s == '#')
		goto free;

	tok->offset = linecur - linestart;

	/* if it's a double quote, wait next double quote */
	if (*s == '"') {
		s++;
		start = s;
		while (*s != '\0') {
			if (*s == '"' && *(s-1) != '\\')
				break;
			s++;
		}
		if (*s == '\0')
			return NULL; /* XXX free */
		linecur = s + 1;
	}
	/* if it's a simple quote, wait next simple quote */
	else if (*s == '\'') {
		s++;
		start = s;
		while (*s != '\0') {
			if (*s == '\'' && *(s-1) != '\\')
				break;
			s++;
		}
		if (*s == '\0')
			goto free;
		linecur = s + 1;
	}
	/* else wait a space */
	else {
		start = s;
		while (*s != '\0' && !isspace(*s)) {
			if (*s == '#')
				break;
			s++;
		}
		linecur = s;
	}
	len = s - start;

	/* allocate string */
	retbuf = strictmalloc(len + 1);
	memcpy(retbuf, start, len);
	retbuf[len] = '\0';

	/* fill token structure and append it in line */
	tok->str = retbuf;
	//printf("Append <%s>\n", tok->str);
	TAILQ_INSERT_TAIL(&line->token_list, tok, next);
	line->n_token ++;

	return linecur;

 free:
	free(tok);
	return NULL;
}

/*
 * Parse the line contained in buffer 'buf'. This function fills a
 * line structure, composed of several tokens.
 */
static struct line *parse_conf_line(const char *linebuf)
{
	struct line *line;
	const char *curbuf = linebuf;

	line = strictmalloc(sizeof(struct line));
	memset(line, 0, sizeof(struct line));
	TAILQ_INIT(&line->token_list);

	if (isspace(*linebuf))
		line->start_with_space = 1;

	/* read tokens from buf and append it to line structure */
	while (curbuf)
		curbuf = append_token(line, linebuf, curbuf);

	return line;
}

int check_opt_name(const char *name)
{
	/* XXX todo */
	(void)name;
	return 0;
}

/* return true if all dependancies are met to set the enable the
 * node */



/* parse a line to get a "help" token */
int parse_help(const struct line *line)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return -1; /* should not happen */

	if (!strcmp(tok->str, "help") && line->n_token == 1)
		return 0;
	else if (!strcmp(tok->str, "---help---") && line->n_token == 1)
		return 0;
	return -1;
}

/* look if given line contains a "source" command, and execute the
 * command if it's the case. Return 0 on success and -1 on error. XXX */
const char *source_file(const struct line *line)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return NULL; /* should not happen */

	/* syntax is: source FILE_NAME */
	if (!strcmp(tok->str, "source") && line->n_token == 2) {
		tok = TAILQ_NEXT(tok, next);
		//printf("source <%s>\n", tok->str);
		return tok->str;
	}
	return NULL;
}

/*
 * Parse the file given as argument and return a tree describing the
 * configuration. XXX wrong
 */
int parse_conf_file(const char *filename, struct confnode **pparent,
		    struct confnode **pcurrent)
{
	char buf[BUFSIZ];
	char *c;
	FILE *f;
	struct line *line;
	struct token *tok;
	int read_help = 0;
	struct confnode *current = *pcurrent;
	struct confnode *parent = *pparent;
	const char *sourcefilename;
	int linenum = 0;

	/* XXX lookup in several PATHs */
	f = fopen(filename, "r");
	if (f == NULL) {
		printf("cannot find file <%s>\n", filename);
		return -1;
	}
	while (fgets(buf, sizeof(buf), f) != NULL) {
		linenum ++;

		/* printf("parent=%s: %s", parent->name, buf); */

		/* we are reading help of an option, do it until the
		 * line does not starts with space */
		if (read_help && (isspace(*buf) || line_is_empty(buf))) {

			/* should not happen */
			if (current == NULL) {
				/* remove \n, display line # */
				c = strchr(buf, '\n');
				if (c != NULL)
					*c = '\0';
				printf("%s:%d invalid line <%s>\n", filename, linenum, buf);
				exit(1);
			}

			/* append string in buffer */
			strncat(current->help, buf,
				sizeof(current->help) - strlen(current->help) - 1);
			continue;
		}

		read_help = 0;

		/* skip empty / comments lines */
		if (line_is_empty(buf))
			continue;

		/* parse line */
		line = parse_conf_line(buf);

		/* XXX put all the following in a function */


		/* if the line does not start with a space, it's a new
		 * keyword, so it's a new node */
		if (!line->start_with_space) {

			/* end of previous node */
			if (current && current->flags & CONFNODE_F_IS_DIR) {
				parent = current;
				current = NULL;
			}

			/* if it's a source command, parse the new
			 * file */
			sourcefilename = source_file(line);
			if (sourcefilename) {
				if (parse_conf_file(sourcefilename, &parent, &current)) {
					/* remove \n, display line # */
					c = strchr(buf, '\n');
					if (c != NULL)
						*c = '\0';
					printf("%s:%d invalid line <%s>\n", filename,
					       linenum, buf);
					exit(1);
				}
				goto free_line;
			}

			/* if the command closes a menu */
			if (confnode_close_dir(parent, line) == 0) {
				parent = parent->parent;
				goto free_line;
			}

			/* if it's a new node command */
			current = confnode_new(line);
			if (current) {
				current->parent = parent;
				TAILQ_INSERT_TAIL(&parent->children, current, next);
				goto free_line;
			}

			/* remove \n, display line # */
			c = strchr(buf, '\n');
			if (c != NULL)
				*c = '\0';
			printf("%s:%d invalid line <%s>\n", filename, linenum, buf);
			exit(1);
		}

		else {
			/* see if it's a help indication */
			if (parse_help(line) == 0) {
				read_help = 1;
				strcpy(current->help, "");
				goto free_line;
			}

			/* parse node attributes */
			if (confnode_add_attr(current, line, buf) == 0)
				goto free_line;

			/* remove \n, display line # */
			c = strchr(buf, '\n');
			if (c != NULL)
				*c = '\0';
			printf("%s:%d invalid line <%s>\n", filename, linenum, buf);
			exit(1);
		}

	free_line:
		while ( (tok = TAILQ_FIRST(&line->token_list)) ) {
			TAILQ_REMOVE(&line->token_list, tok, next);
			free(tok);
		}
		free(line);
	}
	fclose(f);

	*pparent = parent;
	*pcurrent = current;
	/* XXX */
	return 0;
}

/* delete all stored values in conf tree */
void conf_reset(struct confnode *conf)
{
	struct confnode *c;
	conf->value[0] = '\0';
	TAILQ_FOREACH(c, &conf->children, next) {
		conf_reset(c);
	}
}

/*
 * Free previous conf 'prev' if not NULL, then parse the Kconfig and
 * the .config files. Fill the htable. Return the top node of the
 * configuration tree. Return NULL on error.
 */
struct confnode *conf_reset_and_read(struct confnode *prev,
				     const char *dotconfig_filename)
{
	struct confnode *current, *parent;
	struct confnode *root;
	/* XXX free prev */

	confnode_register_all();
	conf_htable_init();
	root = confnode_new_root();
	if (root == NULL)
		return NULL;

	parent = root;
	current = NULL;
#if 1
	parse_conf_file("/home/zer0/projects/libcmdline/Kconfig",
			&parent, &current);
#else
	parse_conf_file("/Users/zer0/projects/libcmdline/Kconfig",
			&parent, &current);
#endif
	conf_htable_fill(root);
	if (dotconfig_read(dotconfig_filename, root) < 0)
		printf("Cannot parse <%s>, use defaults\n", dotconfig_filename);

	return root;
}
