/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "expression.h"
#include "conf_parser.h"
#include "conf_htable.h"
#include "parser_common.h"
#include "dotconfig.h"
#include "confnode.h"

static int confnode_config_new(struct confnode *n, const struct line *line);
static int confnode_config_set_user_strvalue(struct confnode *n,
					      const char *strvalue);
static int confnode_config_get_user_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen);
static int confnode_config_set_default_strvalue(struct confnode *n,
					      const char *strvalue);
static int confnode_config_get_default_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen);
static const char *confnode_config_get_type_str(const struct confnode *n);
static void confnode_config_display_short(const struct confnode *n);

static struct confnode_type confnode_config = {
	.ops = {
		.new = confnode_config_new,
		.free = NULL,
		.add_attr = NULL,
		.close_dir = NULL,
		.dotconfig_write = NULL,
		.strvalue_to_boolvalue = NULL,
		.set_user_strvalue = confnode_config_set_user_strvalue,
		.get_user_strvalue = confnode_config_get_user_strvalue,
		.set_default_strvalue = confnode_config_set_default_strvalue,
		.get_default_strvalue = confnode_config_get_default_strvalue,
		.get_type_str = confnode_config_get_type_str,
		.display_short = confnode_config_display_short,
		.display_long = NULL,
	}
};

/* Create a new node */
static int confnode_config_new(struct confnode *n, const struct line *line)
{
	struct token *tok;
	tok = TAILQ_FIRST(&line->token_list);

	if (strcmp(tok->str, "config") || line->n_token != 2)
		return -1;

	tok = TAILQ_NEXT(tok, next);
	snprintf(n->name, sizeof(n->name), "%s", tok->str);
	n->ops = &confnode_config.ops;
	n->flags = CONFNODE_F_BOOL;
	snprintf(n->default_value, sizeof(n->default_value), "n");
	return 0;
}

/* Set the string value of the node n. Return 0 on success, or -1 if
 * the value cannot be set. */
static int confnode_config_set_user_strvalue(struct confnode *n,
					      const char *strvalue)
{
	if (strcmp(strvalue, "y") && strcmp(strvalue, "n"))
		return -1;
	snprintf(n->value, sizeof(n->value), "%s", strvalue);
	return 0;
}

/* Get the string value that the user asked to set for this node
 * 'n'. This function does not check if dependancies are met. Return 0
 * on success, in this case the string is copied in buffer 'buf' of
 * len 'buflen'. Return -1 if the value cannot be read. */
static int confnode_config_get_user_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen)
{
	if (strcmp(n->value, "") == 0)
		return -1;
	if (!strcmp(n->value, "y")) {
		snprintf(buf, buflen, "y");
		return 0;
	}
	else {
		snprintf(buf, buflen, "n");
		return 0;
	}
}

/* Set the default string value of the node n. Return 0 on success, or
 * -1 if the value cannot be set. */
static int confnode_config_set_default_strvalue(struct confnode *n,
					      const char *strvalue)
{
	if (strcmp(strvalue, "y") && strcmp(strvalue, "n"))
		return -1;
	snprintf(n->default_value, sizeof(n->default_value), "%s", strvalue);
	return 0;
}

/* Get the default string value for this node 'n'. This function does
 * not check if dependancies are met. Return 0 on success, in this
 * case the string is copied in buffer 'buf' of len 'buflen'. Return
 * -1 if the value cannot be read. */
static int confnode_config_get_default_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen)
{
	if (!strcmp(n->default_value, "y")) {
		snprintf(buf, buflen, "y");
		return 0;
	}
	else {
		snprintf(buf, buflen, "n");
		return 0;
	}
}

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *confnode_config_get_type_str(const struct confnode *n)
{
	return "config";
}

/* Print a one-line summary of the node. Used in case of 'ls'. */
static void confnode_config_display_short(const struct confnode *n)
{
	char c;
	if (confnode_get_boolvalue(n) == 0) {
		if (confnode_get_user_boolvalue(n) <= 0)
			c = ' ';
		else
			c = 'd';
	}
	else {
		c = '*';
	}

	printf("[%c]       %s: %s\n", c, n->name, n->prompt);
}

/* register the node type */
void confnode_config_register(void)
{
	TAILQ_INSERT_TAIL(&confnode_type_list, &confnode_config, next);
}
