/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "expression.h"
#include "conf_parser.h"
#include "conf_htable.h"
#include "parser_common.h"
#include "dotconfig.h"
#include "confnode.h"

static int confnode_menu_new(struct confnode *n, const struct line *line);
static int confnode_menu_close_dir(struct confnode *parent,
				   const struct line *line);
static int confnode_menu_dotconfig_write(const struct confnode *n, FILE *f);
static const char *confnode_menu_get_type_str(const struct confnode *n);
static void confnode_menu_display_short(const struct confnode *n);
static void confnode_menu_display_long(const struct confnode *n);

static struct confnode_type confnode_menu = {
	.ops = {
		.new = confnode_menu_new,
		.free = NULL,
		.add_attr = NULL,
		.close_dir = confnode_menu_close_dir,
		.dotconfig_write = confnode_menu_dotconfig_write,
		.strvalue_to_boolvalue = NULL,
		.set_user_strvalue = NULL,
		.get_user_strvalue = NULL,
		.set_default_strvalue = NULL,
		.get_default_strvalue = NULL,
		.get_type_str = confnode_menu_get_type_str,
		.display_short = confnode_menu_display_short,
		.display_long = confnode_menu_display_long,
	}
};

/* Create a new node */
static int confnode_menu_new(struct confnode *n, const struct line *line)
{
	struct token *tok;
	tok = TAILQ_FIRST(&line->token_list);

	if (strcmp(tok->str, "menu") || line->n_token != 2)
		return -1;

	tok = TAILQ_NEXT(tok, next);
	snprintf(n->name, sizeof(n->name), "%s", tok->str);
	n->ops = &confnode_menu.ops;
	n->flags = CONFNODE_F_IS_DIR;
	snprintf(n->default_value, sizeof(n->default_value), "y");
	return 0;
}

/* Parse a line and check if it closes the current node ,
 * "endmenu" closes "menu" node): in this case return 0. Else, return
 * -1. */
static int confnode_menu_close_dir(struct confnode *parent,
				   const struct line *line)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return -1; /* should not happen */

	/* syntax is: endmenu */
	if (!strcmp(tok->str, "endmenu") && line->n_token == 1) {
		//printf("endmenu\n");
		return 0;
	}
	return -1;
}

/* write config value in file f in a dotconfig-like manner. Return 0
 * on success. */
static int confnode_menu_dotconfig_write(const struct confnode *n, FILE *f)
{
	const struct confnode *c;

	if (fprintf(f, "#\n# -- %s\n#\n", n->prompt) < 0)
		return -1;

	TAILQ_FOREACH(c, &n->children, next) {
		if (confnode_dotconfig_write(c, f) < 0)
			return -1;
	}
	return 0;
}

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *confnode_menu_get_type_str(const struct confnode *n)
{
	return "menu";
}

/* Print a one-line summary of the node. Used in case of 'ls'. */
static void confnode_menu_display_short(const struct confnode *n)
{
	printf("       -> %s: %s\n", n->name, n->prompt);
}

/* Print a detailed view of the node. */
static void confnode_menu_display_long(const struct confnode *n)
{
	printf("menu <%s>\n", n->name);
	printf("    path ");
	conf_display_path(n);
	printf("    \"%s\"\n", n->prompt);
}

/* register the node type */
void confnode_menu_register(void)
{
	TAILQ_INSERT_TAIL(&confnode_type_list, &confnode_menu, next);
}
