/*
 * Copyright (c) 2009, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <inttypes.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/queue.h>

#include <cmdline_parse_ipaddr.h>
#include <cmdline_parse.h>

#include "expression.h"
#include "conf_parser.h"
#include "confnode.h"
#include "conf_htable.h"
#include "parse_confnode.h"

/* return 0 if string matches pattern (string must not contain '*')*/
static int
strcmp_joker(const char *s, const char *pattern)
{
	if (s == pattern)
		return 0;

	while (*s == *pattern && *pattern != '*') {
		if (*s == '\0')
			return 0;
		s ++;
		pattern ++;
	}
	if (*pattern != '*')
		return -1;
	if (*(pattern + 1) == '\0')
		return 0;
	while (*s) {
		if (strcmp_joker(s, pattern + 1) == 0)
			return 0;
		s ++;
	}
	return -1;
}

static void
find_in_children(struct confnode_list *l, struct confnode *n,
		 const char *pattern, int flags, int mask, int recurs)
{
	struct confnode *c;

	TAILQ_FOREACH(c, &n->children, next) {
		/* if it's a "if" node, parse children */
		if (c->flags & CONFNODE_F_INVISIBLE)
			find_in_children(l, c, pattern, flags, mask, recurs);

		else {
			if ((c->flags & mask) == flags &&
			    !strcmp_joker(c->name, pattern))
				TAILQ_INSERT_TAIL(l, c, user_next);

			if (recurs)
				find_in_children(l, c, pattern, flags, mask, recurs);
		}
	}
}

static int
parse_conf_node(cmdline_parse_token_hdr_t *tk, const char *buf, void *res,
		unsigned ressize)
{
	struct token_conf_node *tk2 = (struct token_conf_node *)tk;
	struct token_conf_node_data *tkd = &tk2->conf_node_data;
	struct confnode *n;
	struct confnode_list l;
	unsigned int token_len = 0;
	char token[CMDLINE_MAX_TOKEN_SIZE];

	if (res && ressize < sizeof(struct confnode *))
		return -1;

	/* if token is too big... */
	token_len = snprintf(token, sizeof(token), "%s", buf);
	if (token_len >= sizeof(token))
		return -1;

	TAILQ_INIT(&l);

	/* absolute path */
	if (token[0] == '/' && token[1] == '/') {
		/* if contains joker */
		if (strchr(token + 2, '*'))
			find_in_children(&l, *tkd->root, token+2, tkd->flags,
					 tkd->mask, 1);
		else {
			n = conf_htable_lookup(token + 2);
			if ((n->flags & tkd->mask) != tkd->flags)
				return -1;
			TAILQ_INSERT_TAIL(&l, n, user_next);
		}
	}
	/* relative path */
	else
		find_in_children(&l, *tkd->cur, token, tkd->flags, tkd->mask, 0);
	if (TAILQ_EMPTY(&l))
		return -1;

	/* store the address of object in structure */
	if (res)
		*(struct confnode **)res = TAILQ_FIRST(&l);

	return 0;
}

struct confnode_complete_callback {
	struct confnode *start;
	struct confnode *cur;
};


static struct confnode *get_next_node(struct confnode *start,
				      struct confnode *cur, unsigned flags,
				      unsigned mask)
{
	struct confnode *prev;
	struct confnode *parent = start;

	while (1) {
		prev = cur;

		/* no more node to browse */
		if (cur == start)
			return NULL;

		/* first iteration */
		if (cur == NULL)
			cur = TAILQ_FIRST(&parent->children);
		else
			cur = TAILQ_NEXT(cur, next);

		/* no next, return to upper level */
		if (cur == NULL) {
			if (prev == NULL)
				return NULL; /* no node */
			cur = prev->parent;
			continue;
		}

		/* skip invisible nodes, and browse their contents */
		if (cur->flags & CONFNODE_F_INVISIBLE) {
			parent = cur;
			cur = NULL;
			continue;
		}

		/* check that flags match what we want */
		if ((cur->flags & mask) != flags)
			continue;

		/* standard node, return it */
		return cur;
	}
}

static int
complete_conf_node_start(cmdline_parse_token_hdr_t *tk,
			 __attribute__((unused)) const char *tokstr,
			 void **opaque)
{
	struct token_conf_node *tk2 = (struct token_conf_node *)tk;
	struct token_conf_node_data *tkd = &tk2->conf_node_data;
	struct confnode_complete_callback *cb;
	struct confnode *n;
	unsigned flags = tkd->flags;
	unsigned mask = tkd->mask;

	*opaque = NULL;
	cb = malloc(sizeof(*cb));
	if (cb == NULL)
		return -1;
	*opaque = cb;

	n = get_next_node(*tkd->cur, NULL, flags, mask);
	if (n == NULL)
		return -1; /* no completion */

	cb->start = *tkd->cur;
	cb->cur = n;
	return 0;
}

static int
complete_conf_node_iterate(cmdline_parse_token_hdr_t *tk, void **opaque,
			  char *dstbuf, unsigned int size)
{
	struct token_conf_node *tk2 = (struct token_conf_node *)tk;
	struct token_conf_node_data *tkd = &tk2->conf_node_data;
	struct confnode_complete_callback *cb;
	struct confnode *n;
	unsigned len;
	unsigned flags = tkd->flags;
	unsigned mask = tkd->mask;

	cb = *opaque;
	n = cb->cur;
	if (n == NULL)
		return -1;
	cb->cur = get_next_node(cb->start, n, flags, mask);

	len = strlen(n->name) + 1;
	if (len > size)
		return -1;

	strcpy(dstbuf, n->name);
	return 0;
}

static void
complete_conf_node_end(cmdline_parse_token_hdr_t *tk, void **opaque)
{
	if (*opaque)
		free(*opaque);
}

int help_conf_node(cmdline_parse_token_hdr_t *tk, char *dstbuf, unsigned int size)
{
	snprintf(dstbuf, size, "conf-node");
	return 0;
}

struct cmdline_token_ops token_conf_node_ops = {
	.parse = parse_conf_node,
	.complete_start = complete_conf_node_start,
	.complete_iterate = complete_conf_node_iterate,
	.complete_end = complete_conf_node_end,
	.help = help_conf_node,
};

